/***** search.c ***************************************************
 * Description: Search text using keyword tree.
 * Reference: Gusfield, D. (1997). Algorithms on Strings, Trees,
 *    and sequences; Computer Science and Computational Biology.
 *    CUP, p. 60.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Mon Sep  6 22:34:43 2004. 
 *****************************************************************/
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "eprintf.h"
#include "keywordMatcher.h"
#include "stringUtil.h"

/* search a text for a set patterns using the keyword tree root;
 * return the number of hits 
*/
int search(Sequence *patterns, char *text, KNode *root, char *header, int reverse){
  int c, length, found = 0;
  KNode *w, *wdash, *np;

  w = root;
  c = 0;

  length = strlen(text);

  while(c < length){
    while((wdash = findChild(w,text[c])) != NULL){
      w = wdash;
      np = wdash;
      while(np != NULL){
	if(np->id > -1){ 
	  if(found){
	    if(reverse)
	      printf(" %s:%d",patterns[np->id].id,length-c);
	    else
	      printf(" %s:%d",patterns[np->id].id,c-np->depth+2);
	    found++;
	  }else{
	    if(reverse)
	      printf("%s\t%s:%d",header,patterns[np->id].id,length-c);
	    else
	      printf("%s\t%s:%d",header,patterns[np->id].id,c-np->depth+2);
	    found++; 
	  }
	}
	np = np->out;
      }
      c++;
    }
    if(w == root){
      c++;
    }else{
      w = w->fail;
    }
  }
  return found;
}

void searchOnline(Sequence *patterns, char *fileName, KNode *root, Args args){
  FILE *fp;
  char *line, *header, ch;
  int found, k, c;
  KNode *w, *wdash, *np;
  char *delineators = " |\n";

  w = root;
  c = 0;

  line = (char *)malloc(SEQLINE*sizeof(char));
  header = (char *)malloc(SEQLINE*sizeof(char));

  if(fileName == NULL)
    fp = stdin;
  else{
    fp = fopen(fileName, "r");
    if(fp == NULL){
      fprintf(stderr,"error in searchOnline: could not open %s\n.",fileName);
      assert(0);
    }
  }
  line = fgets(line, SEQLINE, fp);
  k=0;
  found = 0;
  while(line != NULL){
      if(args.i)
	ch = tolower(line[k]);     
      else
	ch = line[k];
    if(ch == '>'){
      if(found)
	printf("\n");
      w = root;
      found = 0;
      c = 0;
      /* printf("# Matches in %s",line); */
      header = strncpy(header,strtok(line,delineators),SEQLINE);
      line = fgets(line,SEQLINE,fp);
    }else{
      while((wdash = findChild(w,ch)) != NULL && ch != '>'){
	w = wdash;
	if(wdash->id > -1){ 
	  np = wdash;
	  while(np != NULL){
	    if(found)
	      printf(" %s:%d",patterns[np->id].id,c-np->depth+2);
	    else{
	      printf("%s\t%s:%d",header,patterns[np->id].id,c-np->depth+2);
	      found=1; 
	    }
	    np = np->out;
	  }
	}
	if(line[++k] != '\n'){
	  c++;
	}
	else{
	  line = fgets(line,SEQLINE,fp);
	  if(line == NULL)
	    break;
	  k = 0;
	  c++;
	}
	if(args.i)
	  ch = tolower(line[k]);     
	else
	  ch = line[k];
      }
      if(w == root){
	if(line[++k] != '\n'){
	  c++;
	}else{
	  line = fgets(line,SEQLINE,fp);
	  if(line == NULL)
	    break;
	  k = 0;
	  c++;
	}
      }else{
	w = w->fail;
      }
    }
  }
  printf("\n");
}

void searchSingleSeq(FILE *fp, Args *args, KTree *ktree){
  Sequence *seq, *rev;
  int reverse;
  Sequence *pattern;
  KNode *root;
  
  pattern = ktree->pattern;
  root = ktree->root;

  while((seq=getNextSequence(fp)) != NULL){
    if(args->i){
      seq->seq = strtolower(seq->seq);
    }
    reverse = 0;
    if(search(pattern, seq->seq, root, chomp(seq->id), reverse))
      printf("\n");
    if(args->r){
      reverse = 1;
      rev = revcomp(seq);
      if(args->i){
	rev->seq = strtolower(rev->seq);
      }
      if(search(pattern, rev->seq, root, chomp(rev->id), reverse))
	printf("\n");
      freeSequence(rev);
    }
    freeSequence(seq);
  }
}

void searchMultipleSeq(FILE *fp, Args *args, KTree *ktree){
  Sequence *seqDat, *seq, *rev;
  int fd, i, reverse;
  Sequence *pattern;
  KNode *root;

  pattern = ktree->pattern;
  root = ktree->root;
  seq = (Sequence *)emalloc(sizeof(Sequence));
  fd = fileno(fp);
  seqDat = readMultFasta(fd);

  for(i=0;i<seqDat->numSeq;i++){
    if(i > 0){
      seq->seq = seqDat->seq + seqDat->borders[i-1]+1;
    }else
      seq->seq = seqDat->seq;
    if(args->i)
      seq->seq = strtolower(seq->seq);
    seq->id = seqDat->headers[i];
    reverse = 0;
    if(search(pattern, seq->seq, root, chomp(seq->id), reverse))
      printf("\n");
    if(args->r){
      reverse = 1;
      rev = revcomp(seq);
      if(search(pattern, rev->seq, root, chomp(rev->id), reverse))
	printf("\n");
      freeSequence(rev);
    }
  }
}
