/***** sequenceData.c *********************************************
 * Description: Collection of routines for reading and 
 * manipulating sequence data.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Sun Jun  6 10:34:31 2004.
 *****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include "sequenceData.h"
#include "stringUtil.h"
#include "eprintf.h"

/* everything that is not [acgtACGT] is flagged by a -1 */
int *getRestrictedDnaDictionary(int *dic){
  int i;

  if(dic == NULL)
    dic = (int *)malloc((DICSIZE+1)*sizeof(int));

  for(i=0; i<DICSIZE; i++)
    dic[i] = -1;

  dic['a'] = 0;   /* a */
  dic['c'] = 1;   /* c */
  dic['g'] = 2;   /* g */
  dic['t'] = 3;   /* t */
  dic['A'] = 0;   /* A */
  dic['C'] = 1;   /* C */
  dic['G'] = 2;   /* G */
  dic['T'] = 3;   /* T */

  return dic;

}

/* getDnaDictionary: create DNA dictionary */
int *getDnaDictionary(int *dic){
  int i;

  if(dic == NULL)
    dic = (int *)malloc((DICSIZE+1)*sizeof(int));

  for(i=0; i<DICSIZE; i++)
    dic[i] = 0;

  dic['a'] = 0;   /* a */
  dic['c'] = 1;   /* c */
  dic['g'] = 2;   /* g */
  dic['t'] = 3;   /* t */
  dic['A'] = 0;   /* A */
  dic['C'] = 1;   /* C */
  dic['G'] = 2;   /* G */
  dic['T'] = 3;   /* T */
  dic['r'] = dic['g'];
  dic['R'] = dic['g'];
  dic['y'] = dic['t'];
  dic['Y'] = dic['t'];
  dic['m'] = dic['a'];
  dic['M'] = dic['a'];
  dic['k'] = dic['g'];
  dic['K'] = dic['g'];
  dic['s'] = dic['g'];
  dic['S'] = dic['g'];
  dic['w'] = dic['a'];
  dic['W'] = dic['a'];
  dic['h'] = dic['a'];
  dic['H'] = dic['a'];
  dic['b'] = dic['g'];
  dic['B'] = dic['g'];
  dic['v'] = dic['g'];
  dic['V'] = dic['g'];
  dic['d'] = dic['g'];
  dic['D'] = dic['g'];
  dic['n'] = dic['g'];
  dic['N'] = dic['g'];
  dic['u'] = dic['t'];
  dic['U'] = dic['t'];

  return dic;
}


/* reverse and complement a sequence */
Sequence *revcomp(Sequence *seq){
  int i,j,n;
  char c;
  Sequence *newSeq;

  newSeq = (Sequence *)malloc(sizeof(Sequence));

  n = strlen(seq->seq);
  newSeq->seq = (char *) malloc((n+1)*sizeof(char));
  newSeq->id = strdup2(seq->id);  
  newSeq->id = (char *)realloc(newSeq->id,(strlen(newSeq->id)+11)*sizeof(char));
  strcat(newSeq->id," - REVERSE");
  j=0;
  for(i = n-1; i >= 0; i--){
    c = seq->seq[i];
    switch(c){
    case 'a':
      newSeq->seq[j++] ='t';
      break;
    case 'c':
      newSeq->seq[j++] ='g';
      break;
    case 'g':
      newSeq->seq[j++] ='c';
      break;
    case 't':
      newSeq->seq[j++] ='a';
      break;
    case 'A':
      newSeq->seq[j++] ='T';
      break;
    case 'C':
      newSeq->seq[j++] ='G';
      break;
    case 'G':
      newSeq->seq[j++] ='C';
      break;
    case 'T':
      newSeq->seq[j++] ='A';
      break;
    default:
      newSeq->seq[j++] =c;
      break;
    }
  }
  newSeq->seq[n]='\0';
  return newSeq; 
}

/* Get next sequence from an open data stream in FASTA format; this stream may be the stdin */
Sequence *getNextSequence(FILE *fp){
  static char *buf = NULL;
  static int c = 1;         /* characters in buffer */
  static int bi;            /* buffer index */
  int si;                   /* sequence index */
  int hi;                   /* header index */
  int hs;                   /* maximal header size */
  int ss;                   /* maximal sequence size */
  int headerOpen;
  Sequence *sequence;

  if(c < 1){
    return NULL;
  }

  if(buf == NULL){
    buf = (char *)malloc((BUFSIZ)*sizeof(char));
    c = fread(buf, sizeof(char), BUFSIZ, fp);
    bi = 1;
  }

  /* make a sequence object */
  sequence = (Sequence *) emalloc(sizeof(Sequence));
  hs = BUFSIZ;
  ss = BUFSIZ;
  /* allocate memory for sequence id */
  sequence->id = (char *)emalloc(hs*sizeof(char));
  /* allocate memory for sequence string */
  sequence->seq = (char *)emalloc(ss*sizeof(char));

  ss = BUFSIZ;
  si = 0;
  hi = 0;
  headerOpen = 1;
  while(bi < c){
    if(buf[bi] == '>'){
      bi++;
      sequence->seq = (char *) erealloc(sequence->seq,(si+1)*sizeof(char));
      sequence->seq[si] = '\0';
      si = 0;
      sequence->id = (char *)erealloc(sequence->id,(hi+1)*sizeof(char));
      sequence->id[hi] = '\0';
      hi = 0;
      headerOpen = 1;
      return sequence;
    }
    if(si == ss){
      ss *= 2;
      sequence->seq = (char *)erealloc(sequence->seq,ss*sizeof(char));
    }
    if(hi == hs){
      hs *= 2;
      sequence->id = (char *)erealloc(sequence->id,hs*sizeof(char));
    }
    if(headerOpen){
      if(buf[bi] != '\n')
	sequence->id[hi++] = buf[bi++];
      else{
	headerOpen = 0;
	bi++;
      }
    }else{
      if(buf[bi] != '\n')
	sequence->seq[si++] = buf[bi++];
      else
	bi++;
    }
    if(bi == c){
      c = fread(buf,sizeof(char),BUFSIZ,fp);
      bi = 0;
    }
  }
  return sequence;
}

/* read FASTA file and store data; return size of data set */
Sequence *readFasta(char *file, Sequence *data, int *size, int *residues){
    FILE *fp;
    char *line;
    int first, l, numentry, maxseqlen, seqlen;
    int *lengthptr;
    Sequence *od;

    line = (char *)malloc((SEQLINE+2)*sizeof(char));

    /* open input file NB: this has got to be an external file, otherwise the rewind fails */
    fp = fopen(file, "r");
    if(fp == NULL){
	fprintf(stderr,"error in readFasta: could not open %s\n.",file);
	exit(1);
    }
    
    /* count number of entries and determine length of longest sequence */
    numentry = 0;
    maxseqlen = 0;
    seqlen = 0;
    *residues = 0;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strlen(line)>SEQLINE){
	    printf("error in readFasta: cannot deal with lines longer than %d bp.\n",SEQLINE);
	    printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
	    exit(2);
	}
	if(strstr(line,">") != NULL){
	    numentry++;
	    if(seqlen > maxseqlen)
		maxseqlen = seqlen;
	    seqlen = 0;
	}else{
	    l = strlen(line) - 1;
	    seqlen += l; 
	    *residues += l;
	}
    }
    *size = numentry;
    /* allocate memory */
    data = (Sequence *)malloc(numentry*sizeof(Sequence));
    lengthptr = (int *)malloc(numentry*sizeof(int));
    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    od = data;
    rewind(fp);
    /* Determine the length of each input sequence */
    lengthptr--;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strstr(line,">") != NULL){
	    lengthptr++;
	    *lengthptr=0;
	}else{
	    l = strlen(line) - 1;
	    *lengthptr += l;
	}
    }

    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    rewind(fp);

    first = 1;
    lengthptr -= numentry;
    data--;
    while((line=fgets(line,SEQLINE,fp)) != NULL){
	line = chomp(line);
	if(strstr(line,">") != NULL){           /* found new entry */
	    if(first == 0){
		data->seq -= *lengthptr;        /* readjust sequence pointer position */
	    }else{
		first = 0;
	    }
	    lengthptr++;
	    data++;
	    data->seq = (char *)malloc((*lengthptr + 1) * sizeof(char));
	    data->id = (char *)malloc((strlen(line)+1)*sizeof(char));
	    data->id = strdup2(line+1);
	}else{
	    strcpy(data->seq,line);
	    data->seq += strlen(line);
	}
    }
    /* readjust last sequence pointer */
    data->seq -= *lengthptr;
    free(line);
    return od;
}

/* read FASTA-formatted sequence data from an open file descriptor */
Sequence *readMultFasta(int fd){
  Sequence *s;
  char *buf;
  int headerOpen;
  int headerLen;
  long i, maxLen, c;

  buf = (char *)emalloc(BUFSIZ*sizeof(char));
  s = (Sequence *)emalloc(sizeof(Sequence));
  s->id = emalloc(6*sizeof(char));
  s->id[0] = '\0';
  strcat(s->id,"strId");
  s->seq = (char *)emalloc(0*sizeof(char));
  s->borders = (int *)emalloc(0*sizeof(int));
  s->headers = (char **)emalloc(0*sizeof(char *));
  headerOpen = 0;
  s->len = 0;
  s->numSeq = 0;
  maxLen = 0;
  headerLen = 0;
  while((c = read(fd,buf,BUFSIZ)) > 0){
    if(s->len + c > maxLen){
      if(maxLen >= BUFSIZ)
	maxLen *= 2;
      else
	maxLen = BUFSIZ + 1;
      s->seq = (char *)erealloc(s->seq,maxLen*sizeof(char));
    }
    for(i=0; i<c; i++){
      if(buf[i] == '>'){
	headerOpen = 1;
	/* take care of sequence borders */
	s->borders = (int *)erealloc(s->borders,(s->numSeq+1)*sizeof(int));
	if(s->numSeq >= 1){
	  s->seq[s->len] = '\0';  /* unique character to terminate each sequence */
	  s->len++;
	  s->borders[s->numSeq-1] = s->len-1;
	}
	/* take care of sequence headers */
	s->headers = (char **)erealloc(s->headers,(s->numSeq+1)*sizeof(char *));
	s->headers[s->numSeq] = (char *)emalloc(BUFSIZ*sizeof(char));
	headerLen = 0;
	s->numSeq++;
      }
      if(headerOpen){
	if(buf[i] == '\n'){
	  headerOpen = 0;
	  s->headers[s->numSeq-1][headerLen] = '\0';
	  s->headers[s->numSeq-1] = (char *)erealloc(s->headers[s->numSeq-1],(headerLen+1)*sizeof(char));
	}
	else if(headerLen < BUFSIZ && isprint(buf[i]))
	  s->headers[s->numSeq-1][headerLen++] = buf[i];
      }else if(buf[i] != '\n'){
	s->seq[s->len] = buf[i];
	s->len++;
      }
    }
  }
  /* add last border */
  s->seq[s->len] = '\0';
  s->len++;
  /* set end of last sequence read */
  s->borders[s->numSeq - 1] = s->len - 1;
  /* trim sequence to actual size */
  s->seq = (char *)erealloc(s->seq,s->len*sizeof(char));
  return s;
}

/* freeSequence: free the data structure Sequence */
Sequence *freeSequence(Sequence *seq){
  free(seq->id);
  free(seq->seq);
  free(seq);
  return seq;
}
